// 1825ServoTester.c   Servo Tester with 4D Systems Display
//
// PIC:           16F1825 - 14-pin
//                Pullups on all pins
// Compiler:      CCS PCWH 5.048
// Programmer:    PICkit2
// Hardware:      4D Systems Display uLCD-24PTU - runs at 5V, has 3.3V out, if needed
//                CCP Servo output - via timers, not PWM module
//                Battery voltage measured via voltage divider
//                
//                   +5V VDD - 1   14 - VSS -
//  HB LED    <-    OSC1/RA5 - 2   13 - RA0/PGD     
//                  OSC2/RA4 - 3   12 - RA1/PGC
//                  MCLR/RA3 - 4   11 - RA2/CCP3          -> RTC check pin
//  4D out    -> CCP1/RX/RC5 - 5   10 - RC0/SCL/SCK
//  4D in     <-      TX/RC4 - 6    9 - RC1/SDA/SDI/CCP4  <- AN5 Voltage in from divider
//  Servo PWM <-    CCP2/RC3 - 7    8 - RC2/    SDO       -> JGSDBG debug output
//
//  Servo code from: www.ccsinfo.com/forum/viewtopic.php?t=44328
//
//
// 2015.08.11 - Jurgen G Schmidt               www.jgscraft.com
// 2015.08.25 - added voltage detector
// 2015.08.29 - all working, comments cleanup
//
//------------------------------------- 4D Systems IDE Display Notes ---------------
//  List the display variables here...
//
//  ------ Writing to the display ------ all values hex ------
//
//  Strings0   02 11 00 len data  - version ID string
//  Strings1   02 11 01 len data  - current servo pulse value as string
//  Strings2   02 11 02 len data  - Green voltage display
//  Strings3   02 11 03 len data  - Red voltage display
//  Leddigits0 01 0F 01 msb lsb   - current servo pulse value
//  Slider0    01 04 01 msb lsb   - current servo pulse value
//
//  ------ Received Data ------
//
//  Winbutton0 07 06 00 00  00    - Decr button (-10)
//  Winbutton1 07 06 01 00  00    - Incr button (+10)
//  Winbutton2 07 06 02 00  00    - Incr button (-1)
//  Winbutton3 07 06 03 00  00    - Incr button (+1)
//  Winbutton4 07 06 04 00  00    - Min (1000)
//  Winbutton5 07 06 05 00  00    - Mid (1500)
//  Winbutton5 07 06 06 00  00    - Max (2000)
//  Slider0    07 04 00 msb lsb   - Slider
//
//------------------------------------- HARDWARE -------------------------------
#include <16F1825.h>
#device ansi
#device ADC=10        //hi res ADC

#FUSES NOWDT          //No Watch Dog Timer
#FUSES NOPUT          //No Power Up Timer
#FUSES NOMCLR         //No Reset Button
#FUSES NOBROWNOUT     //No Brownout Reset

#use delay( internal=32M )

// Serial debug data via SW. INVERT so we can connect directly to legacy PC serial port
// (Usually via USB converter)
#use rs232( BAUD=19200, XMIT=PIN_C2, INVERT, STREAM=JGSDBG )

// Be sure to set matching baud rate in the 4DS IDE
#use rs232( BAUD=38400, UART1, STREAM=DISPLAY, ERRORS )

#define LED_HB      PIN_A5    // heartbeat LED

#define SERVO_PIN   PIN_C3    // servo control signal

typedef unsigned int8   UINT8;
typedef unsigned int16  UINT16;

#include  <string.h>

//------ My universal timer stuff...
//-------------------------------------- int_timer4 ----------------------------
// Tick timer
//
//  24hrs = 1440min = 86,400secs
//  65,536mins = 45days rollover
//
UINT16  intMSecs=0;     // milliseconds counter
UINT16  intSecs=0;      // seconds counter
UINT16  intOldSecs=0;   // checks for seconds rollover
UINT16  intMins=0;      // minutes counter
UINT16  intOldMins=0;   // checks for minutes rollover

#int_timer4
void SecTick()
{
  intMSecs++;
  //------ increment millisecond timers here
   
  if( intMSecs > 900L )                 // heartbeat LED
    output_high( LED_HB );
  else
    output_low( LED_HB );

  if( intMSecs > 999L ) {
    intMSecs=0;
    intSecs++;
    //------ increment second timers here

  }
   
  if( intSecs > 59 ) {
    intSecs = 0;   
    intMins++;
    //------ increment minute timers here

  }
  // output_toggle( PIN_A2 );  // connect to o-scope for timer calibration
}

//-----------------------------------------------------------------------------
//------------------------------------- Servo Driver--- -----------------------

//------------------------------------- #int_ccp1 -----------------------------
// These pulse duration values are for a 32 MHz PIC oscillator
// frequency and a Timer1 divisor of 8.
// Results in Timer1 counting microseconds. When timer1 matches CCP_1 
// we get an interrupt and the interrupt code executes. We just keep adding the 
// new wait time to CCP_1 and it rolls over just like the timer does. Neat!
// We could do this easier with the PWM module but it doesn't go slow enough.
#define PWM_PERIOD 20000L  // Gives 20ms PWM period (50 Hz)
#define PULSE_1MS   1000L  // Gives 1ms high level pulse
#define PULSE_2MS   2000L  // Gives 2ms high level pulse
#define PULSE_MID   1500L  // supposed center position

UINT16  pulse_high_duration = PULSE_MID;  // initialize centered
UINT8   pulse_done_flag;

//-----------------------------------
#int_ccp1
void ccp1_isr(void)
{
  static int8 set_servo_pin_high = TRUE;

  // If servo pin low level pulse time is finished,
  // then do the high portion of the signal.
  if(set_servo_pin_high) 
  {
     output_high(SERVO_PIN);     
     set_servo_pin_high = FALSE;
     CCP_1 += pulse_high_duration;   
  } else {
    // If high portion of signal is done, do the low part.
    output_low(SERVO_PIN);       
    set_servo_pin_high = TRUE;
    CCP_1 += (PWM_PERIOD - pulse_high_duration);   
   
    pulse_done_flag = TRUE;
  }
}

//-----------------------------------------------------------------------------
//------------------------------------- 4D Display Stuff ----------------------

//------------------------------------- #int_rda ------------------------------
//-- process data received from display

#define  DS_BUFFER_SIZE 64              // define buffer size
BYTE     strDSbuffer[DS_BUFFER_SIZE];   // allocate buffer
UINT8    nextIn = 0;                    // last received position - set by ISR
UINT8    nextOut = 0;                   // last read by application - set by bgetc()

#int_rda
void serial_isr() {

   UINT8 t;

   strDSbuffer[nextIn]=fgetc(DISPLAY);
   //fprintf(JGSDBG, "%2x^", strDSbuffer[nextIn]);
   t=nextIn;
   nextIn=(nextIn+1) % DS_BUFFER_SIZE;
   if(nextIn==nextOut)
     nextIn=t;           // Buffer full !!
}

#define bkbhit (nextIn!=nextOut)

BYTE bgetc() {
   BYTE c;

   while(!bkbhit) ;
   c=strDSbuffer[nextOut];
   nextOut = (nextOut+1) % DS_BUFFER_SIZE;
   return(c);
}

BYTE    strDScmd[DS_BUFFER_SIZE];       // received data 
UINT8   intDSindex;                     // index to received data
BYTE    chDSgot;                        // character we got from strDSbuffer

#define DS_ACK      0x06  // ACK - display received OK
#define DS_NAK      0x15  // NAK - display reports bad data received
#define OFF         0
#define ON          1 

BYTE    statusDS ;        // last display status message ACK or NAK
UINT16  intDSServo;       // servo setting from slider


//------ Objects can be stacked in the 4D display. Last one to receive a message goes to top

BYTE  strDSwork[ 64 ];    // assemble message here, first element is always length before checksum

// static commands - 1st byte is length, do not include checksum
//BYTE  strDSoff[]    = { 0x02, 0x04, 0x00 };
//BYTE  strDSdim[]    = { 0x02, 0x04, 0x07 };
BYTE  strDSbright[] = { 0x02, 0x04, 0x0f };
BYTE  strDSclick[]  = { 0x05, 0x01, 0x16, 0x00, 0x00, 0x00 };   // play key click sound
BYTE  strDSform0[]  = { 0x05, 0x01, 0x0a, 0x00, 0x00, 0x00 };   // Form0 - user code


//------------------------------------- DS_Send() -----------------------------
//-- Send data to display and append check digit
//-- first byte is length of the data to send - might have embedded 0x00 so
//-- can't rely on NULL termination
//
void DS_Send( BYTE data[] )
{
  UINT8 len;          // save length here for readability
  UINT8 i;            // loop counter
  BYTE  chk=0;        // XOR checksum
  
  len = data[0];
  for( i=1; i<=len; i++ ) {             // skip length byte
    fputc( data[i], DISPLAY );          // send data
    chk ^= data[i];                     // cumulative XOR
  }
  fputc( chk, DISPLAY );                // send checksum
}

//------------------------------------- DS_SetDigits() ------------------------
//-- Set LED Digits
//
void DS_SetLed( UINT8 idx, UINT16 value )
{
  strDSwork[0] = 0x05;                  // length is always 5 (before checksum)
  strDSwork[1] = 0x01;                  // WRITE_OBJ
  strDSwork[2] = 0x0F;                  // leddigits
  strDSwork[3] = idx;                   // which one
  strDSwork[4] = make8( value, 1 );     // get MSB
  strDSwork[5] = make8( value, 0 );     // get LSB
  
  DS_Send( strDSwork );                 // send it !
}

//------------------------------------- DS_SetSlider() ------------------------
//-- Set Slider Value
//
void DS_SetSlider( UINT8 idx, UINT16 value )
{
  strDSwork[0] = 0x05;                  // length is always 5 (before checksum)
  strDSwork[1] = 0x01;                  // WRITE_OBJ
  strDSwork[2] = 0x04;                  // slider
  strDSwork[3] = idx;                   // which one
  strDSwork[4] = make8( value, 1 );     // get MSB
  strDSwork[5] = make8( value, 0 );     // get LSB
  
  DS_Send( strDSwork );                 // send it !
}

//------------------------------------- DS_SetString() ------------------------
//-- Send string data to display and append check digit
//
void DS_SetString( UINT8 idx, BYTE str[] )
{
  UINT8 len;
  
  len = strlen( str );
  strDSwork[0] = len + 4;
  strDSwork[1] = 0x02;                  // WRITE_STR
  strDSwork[2] = idx;                   // string index
  strDSwork[3] = len + 1;               // #chars + terminator
  strcpy( strDSwork+4, str );
  
  DS_Send( strDSwork );
}

//-----------------------------------------------------------------------------
void TestServo( void );
//-----------------------------------------------------------------------------
//------------------------------------- main() --------------------------------

void main()
{
  BYTE    strWork[64];
  UINT16  intVolts, intVunits, intVdecimals;

  setup_adc_ports( NO_ANALOGS);         // ALWAYS, Always, always do this !!!

  // Timer2,4,6,8,10 has prescalar 1/4/16 and post 1-16
  // 32MHz -> 8MHz -> .125us per cycle
  // timer rollover = .125 x 200 (0xC8) = 25us per timer click
  // T2_DIV_BY_4 = 25us x 4 = 100us per timer click   
  // Postscalar of 10 gives 1mS tick
  //
  setup_timer_4( T4_DIV_BY_4, 0xC8, 10 );     // adjust second parm as needed via scope

  setup_timer_1( T1_DIV_BY_8 | T1_INTERNAL ); // 1 us tick for servo CCP

  // Setup CCP to generate CCP1 interrupt when Timer1
  // matches the value in the CCP1 registers.
  setup_ccp1(CCP_COMPARE_INT);
  output_low(SERVO_PIN);
  
  // Initialize the variables and CCP1 to center servo position.
  pulse_high_duration = PULSE_MID; 
  CCP_1 = PWM_PERIOD - pulse_high_duration;  // Pulse Off time
  
  // This flag tells the main loop code (below) when it's OK
  // to change the pulse_high_duration variable.
  pulse_done_flag = FALSE;

  setup_adc_ports( sAN5 );
  setup_adc( ADC_CLOCK_INTERNAL );

  delay_ms( 1500 );                      // allow display to wake up :)
  fprintf(JGSDBG, "\r%s %s %s\r\n", __FILENAME__, __DATE__, __TIME__);
  sprintf(strWork, "  %s %s %s ", __FILENAME__, __DATE__, __TIME__);

  clear_interrupt(INT_CCP1);
  enable_interrupts(INT_CCP1);
  enable_interrupts(INT_RDA);
  enable_interrupts(INT_TIMER4);
  enable_interrupts(GLOBAL);

  // Display initialize
  DS_Send( strDSclick );
  delay_ms( 100 );
  DS_Send( strDSclick );
  DS_Send( strDSbright );
  DS_Send( strDSform0 );                // main screen
  DS_SetString( 0, strWork );
  
  delay_ms( 500 );
  intDSServo = pulse_high_duration;
  sprintf( strWork,"   %lu uS", intDSServo);
  DS_SetString( 1, strWork );
  DS_SetSlider( 0, intDSServo );
  DS_SetLed( 0, intDSServo );

  //----------------------------------- BIG LOOP ------------------------------

  //TestServo();  // test servo code - comment out to run rest of program

  while( TRUE ){

    //-------------------------------------------------------------------------
    //------ Check for and process data received from display - 1 or 6 bytes 
    //
    while( bkbhit ) {
      chDSgot = bgetc();
      //fprintf( JGSDBG,"[%02x]", chDSgot );
      if( (chDSgot==DS_ACK || chDSgot==DS_NAK) && intDSindex==0 ) {
        statusDS = chDSgot;             // latest status
        fprintf( JGSDBG, "\r!%02x# ", statusDS );
      } else {
        strDScmd[intDSindex++] = chDSgot;
      }
    }  
    
    if( intDSindex>5 ) {                // process data received from display
      
      fprintf( JGSDBG, "\r>%02x %02x %02x %02x %02x %02x <", strDScmd[0], strDScmd[1], strDScmd[2], strDScmd[3], strDScmd[4], strDScmd[5]);
      DS_Send( strDSclick );


      //-- decode screen commands
      //
      if( strDScmd[1]==0x04 ) {                       //--- Slider
        intDSServo = strDScmd[3]*256L + (UINT16)strDScmd[4];
        fprintf( JGSDBG,"--%lu uS  %u %u --", intDSServo, strDScmd[3], strDScmd[4]);
      }
      
      if( strDScmd[1]==0x06 && strDScmd[2]==0x00 ) {  //--- Winbutton0  -10
        fputs( "btn0 - ", JGSDBG );
        intDSServo -= 10L;
        if( intDSServo < 100L ) intDSServo = 100L;
      }
      
      if( strDScmd[1]==0x06 && strDScmd[2]==0x01 ) {  //--- Winbutton1  +10
        fputs( "btn1 + ", JGSDBG );
        intDSServo += 10L;
        if( intDSServo > 2500L ) intDSServo = 2500L;
      }

      if( strDScmd[1]==0x06 && strDScmd[2]==0x02 ) {  //--- Winbutton2  -1
        fputs( "btn2 - ", JGSDBG );
        intDSServo -= 1L;
        if( intDSServo < 100L ) intDSServo = 100L;
      }
      
      if( strDScmd[1]==0x06 && strDScmd[2]==0x03 ) {  //--- Winbutton3  +1
        fputs( "btn3 + ", JGSDBG );
        intDSServo += 1L;
        if( intDSServo > 2500L ) intDSServo = 2500L;
      }

      if( strDScmd[1]==0x06 && strDScmd[2]==0x04 ) {  //--- Winbutton4  MIN
        fputs( "btn4 + ", JGSDBG );
        intDSServo = 1000L;
      }

      if( strDScmd[1]==0x06 && strDScmd[2]==0x05 ) {  //--- Winbutton5  MID
        fputs( "btn5 + ", JGSDBG );
        intDSServo = 1500L;
      }

      if( strDScmd[1]==0x06 && strDScmd[2]==0x06 ) {  //--- Winbutton6  MAX
        fputs( "btn6 + ", JGSDBG );
        intDSServo = 2000L;
      }

      //-- display pulse width in microseconds
      sprintf( strWork,"   %lu uS", intDSServo);
      DS_SetString( 1, strWork );
      DS_SetSlider( 0, intDSServo );
      DS_SetLed( 0, intDSServo );

      //-- update servo signal with pulse width
      while(!pulse_done_flag);    // Wait until pulse is done, then update
      pulse_done_flag = FALSE;
      disable_interrupts(GLOBAL);
      pulse_high_duration = intDSServo;   
      enable_interrupts(GLOBAL);
      
      intDSindex = 0;
    }


    //-------------------------------------------------------------------------
    //------ do other stuff here ----------------------------------------------

    //-- stuff to do every second --
    
    if( intOldSecs != intSecs ) {
      
      //-------------------------------- Voltage Display ----------------------
      //--- Battery voltage goes through a voltage divider to ADC pin
      //--- ADC range is 0-1023, for 0-5V
      //--- Since we divided the voltage, our reading matches pretty closely,
      //--- save for shifting the decimal point. We're only interested in 
      //--- approximate voltage so we know when to recharge the batteries.
      
      set_adc_channel( 5 );             // slect our ADC channel
      delay_us( 10 );                   // let it settle
      intVolts = read_adc();            // get a reading

      intVunits = intVolts / 100;       // get the unit volts
      intVdecimals = intVolts % 100;    // get decimals
      
//      fprintf( JGSDBG, "\rVolts: %lu  %lu.%lu \r", intVolts, intVunits, intVdecimals );
      
      //-- display voltage - take advantage of stacked objects to show different colors,
      //-- since we can't pass a color parameter. 
      //-- Whichever one gets displayed last is on top.
      sprintf( strWork,"%lu.%lu V", intVunits, intVdecimals );
      if( intVunits < 6 )
        DS_SetString( 3, strWork );     // red version
      else
        DS_SetString( 2, strWork );     // green version

      intOldSecs = intSecs;
    }

    //-- stuff to do every minute --
    
    if( intOldMins != intMins ) {

      intOldMins = intMins;
    }
    
    delay_ms( 1 );                      // slow down a little
    output_toggle( PIN_A4 );            // check loop time

  }//end of BIG LOOP

}//end of main()

//-----------------------------------------------------------------------------
//  TestServo - this code verifies that I've done the timer calculations correctly
//  - connect oscilloscope to servo pin and you should see a wavelength of 20ms
//  - there should be a square-wave peak that widens and narrows between 1ms and 2ms
//
void TestServo()
{
  UINT16  i;
  
  while(TRUE)
    {
     // Slowly increase the pulse from 1ms to 2ms.
     for(i = PULSE_1MS; i < PULSE_2MS; i+=4)
        {
         while(!pulse_done_flag); // Wait until pulse is done
  
         pulse_done_flag = FALSE;
         disable_interrupts(GLOBAL);
         pulse_high_duration = i;   
         enable_interrupts(GLOBAL);
        }
  
     for(i = PULSE_2MS; i > PULSE_1MS; i-=4)
        {
         while(!pulse_done_flag); // Wait until pulse is done
  
         pulse_done_flag = FALSE;
         disable_interrupts(GLOBAL);
         pulse_high_duration = i;   
         enable_interrupts(GLOBAL);
        }
    }
}

//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
//------ end of 1825ServoTester.c ------
